from datetime import date

import pytest
from pydicom import Dataset
from pydicom.errors import InvalidDicomError

from .base_client import BaseDicomClient, PRIVATE_ID


def dataset_factory(defaults):
    '''
    Create a factory function for building pydicom datasets.

    If any values are functions, they will be called to determine the value of
    that attribute.  When the function is called, it will be passed an integer
    indicating the number of times the factory has been called, and the full
    attribute dict.

    If a value is `None`, then it is not set.

    :param defaults: Dict whose keys are tag names and values are tag values or functions.
    :return: A factory function that produces datasets.  Keyword arguments
        override the defaults.
    '''
    counter = 0

    def factory(**overrides):
        nonlocal counter
        ds = Dataset()
        attributes = {**defaults, **overrides}
        for key, value in attributes.items():
            if callable(value):
                value = value(counter, attributes)
            if value is not None:
                setattr(ds, key, value)
        counter += 1
        return ds
    return factory


@pytest.fixture
def patient_dataset_factory():
    defaults = {
        'PatientID': 'PAT001',
        'PatientName': 'John^Doe',
        'PatientBirthDate': date(2000, 1, 1),
        'StudyInstanceUID': '100000',
        'StudyDate': date(2018, 1, 1),
        'SOPInstanceUID': lambda i, a: f'1{i:05}',
    }
    return dataset_factory(defaults)


update_patient_result = BaseDicomClient.update_patient_result


def test_update_patient_result_single_slice(patient_dataset_factory):
    slice_dataset = patient_dataset_factory()
    result = Dataset()
    update_patient_result(result, slice_dataset)
    assert result.PatientID == slice_dataset.PatientID
    assert result.PatientName == slice_dataset.PatientName
    assert result.PacsmanPrivateIdentifier == PRIVATE_ID
    assert len(result.PatientStudyInstanceUIDs) == 1
    assert result.PatientStudyInstanceUIDs[0].name == slice_dataset.StudyInstanceUID.name
    assert result.PatientMostRecentStudyDate == slice_dataset.StudyDate


def test_update_patient_result_raise_if_id_change(patient_dataset_factory):
    slice_1 = patient_dataset_factory(PatientID='1')
    slice_2 = patient_dataset_factory(PatientID='2')
    result = Dataset()
    update_patient_result(result, slice_1)
    with pytest.raises(ValueError):
        update_patient_result(result, slice_2)


def test_update_patient_result_no_raise_if_name_change(patient_dataset_factory):
    '''
    At the moment, we grab values from the first dataset that we see, and we
    don't change any of them after that.  This is probably the desired
    behaviour, but it may be worth considering grabbing details from the most
    recent DICOM dataset
    '''
    result = Dataset()
    update_patient_result(result, patient_dataset_factory(PatientName='1'))
    update_patient_result(result, patient_dataset_factory(PatientName='2'))
    assert result.PatientName == '1'


def test_update_patient_result_multiple_studys(patient_dataset_factory):
    result = Dataset()
    update_patient_result(result, patient_dataset_factory(StudyInstanceUID='1'))
    update_patient_result(result, patient_dataset_factory(StudyInstanceUID='2'))
    assert len(result.PatientStudyInstanceUIDs) == 2
    assert {uid.name for uid in result.PatientStudyInstanceUIDs} == {'1', '2'}


def test_update_patient_result_single_study(patient_dataset_factory):
    result = Dataset()
    update_patient_result(result, patient_dataset_factory(StudyInstanceUID='1'))
    update_patient_result(result, patient_dataset_factory(StudyInstanceUID='1'))
    assert len(result.PatientStudyInstanceUIDs) == 1
    assert result.PatientStudyInstanceUIDs[0].name == '1'


def test_update_patient_result_most_recent_study_date(patient_dataset_factory):
    result = Dataset()
    update_patient_result(result, patient_dataset_factory(StudyDate=date(2018, 1, 1)))
    assert result.PatientMostRecentStudyDate == date(2018, 1, 1)
    update_patient_result(result, patient_dataset_factory(StudyDate=date(2018, 1, 2)))
    assert result.PatientMostRecentStudyDate == date(2018, 1, 2)
    update_patient_result(result, patient_dataset_factory(StudyDate=date(2018, 1, 1)))
    assert result.PatientMostRecentStudyDate == date(2018, 1, 2)


def test_update_patient_result_missing_study_date(patient_dataset_factory):
    result = Dataset()
    update_patient_result(result, patient_dataset_factory(StudyDate=''))
    assert result.PatientMostRecentStudyDate == ''
    update_patient_result(result, patient_dataset_factory(StudyDate=date(2018, 1, 1)))
    assert result.PatientMostRecentStudyDate == date(2018, 1, 1)


def test_update_patient_result_additional_tags_present(patient_dataset_factory):
    result = Dataset()
    dataset = patient_dataset_factory(PatientSex='M')
    additional_tags = ['PatientSex']
    update_patient_result(result, dataset, additional_tags)
    assert result.PatientSex == 'M'


def test_update_patient_result_additional_tags_absent(patient_dataset_factory):
    result = Dataset()
    dataset = patient_dataset_factory()
    additional_tags = ['PatientSex']
    update_patient_result(result, dataset, additional_tags)
    assert result.PatientSex == ''


@pytest.mark.parametrize('attribute', [
    'PatientID',
    'StudyInstanceUID',
])
def test_update_patient_result_unhandled_missing_tags(patient_dataset_factory, attribute):
    overrides = {attribute: None}
    with pytest.raises(InvalidDicomError):
        update_patient_result(Dataset(), patient_dataset_factory(**overrides))


@pytest.mark.parametrize('attribute', [
    'PatientName',
    'PatientBirthDate',
    'StudyDate',
])
def test_update_patient_result_handled_missing_tags(patient_dataset_factory, attribute):
    overrides = {attribute: None}
    update_patient_result(Dataset(), patient_dataset_factory(**overrides))


@pytest.mark.parametrize('attribute', [
    'PatientName',
    'PatientBirthDate',
    'StudyDate',
])
def test_update_patient_result_empty_tags(patient_dataset_factory, attribute):
    overrides = {attribute: ''}
    update_patient_result(Dataset(), patient_dataset_factory(**overrides))



import os
import argparse

from pacsman import PynetDicomClient
from pacsman import dicom_file_iterator

TEST_DATA_DIRECTORY = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test_dicom_data')
LOCAL_PACS_URL = os.environ.get('LOCAL_PACS_URL', 'localhost')


def main(destination='both'):
    print(f'uploading test data from {TEST_DATA_DIRECTORY}')

    if destination == 'both' or destination == 'remote':
        print('Uploading files to remote destination')
        client = PynetDicomClient(client_ae='TEST', pacs_url=LOCAL_PACS_URL, pacs_port=4242,
                                  dicom_dir='.', remote_ae='ORTHANC')
        client.send_datasets(dicom_file_iterator(TEST_DATA_DIRECTORY))

    if destination == 'both' or destination == 'local':
        print('Uploading files to local destination')
        # "TEST" must be registered AE in the Horos listener
        client = PynetDicomClient(client_ae='TEST', pacs_url=LOCAL_PACS_URL, pacs_port=11112,
                                  dicom_dir='.', remote_ae='HOROS-LOCAL')
        client.send_datasets(dicom_file_iterator(TEST_DATA_DIRECTORY))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('destination', choices=['both', 'remote', 'local'], default='both', nargs='?')
    args = parser.parse_args()
    main(destination=args.destination)



'''
This filesystem client can be used for testing in development when a PACS server
is not available. It may be slow if many datasets are present: All get/fetch operations
are O(N) on the number of DICOM datasets loaded from the `test_dicom_data` dir.

Example data located in `test_dicom_data` dir:
 (from www.dicomserver.co.uk).

Patient ID PAT001 "Joe Bloggss" dob 19450703
    Study ID 1.2.826.0.1.3680043.11.1011
        Study Date 20180522
        CT Modality
            Series ID 1.2.826.0.1.3680043.6.86796.74495.20180522152336.14136.1.23
                1 Image

Patient ID PAT014 "Erica Richardson" dob 19520314
    Study ID 1.2.826.0.1.3680043.11.118
        Study Date 20180518
        CT Modality
            Series ID 1.2.826.0.1.3680043.6.51581.36765.20180518132103.25992.1.21
                5 images
    Study ID 1.2.826.0.1.3680043.11.118.1
'''
import glob
import logging
import os
import shutil
from datetime import datetime
from collections import defaultdict
from typing import List, Optional, Dict, Iterable

from pydicom import dcmread, Dataset
from pydicom.valuerep import MultiValue
from pydicom.uid import UID

from .base_client import BaseDicomClient, PRIVATE_ID
from .utils import process_and_write_png_from_file, copy_dicom_attributes, dicom_filename

logger = logging.getLogger(__name__)


class FilesystemDicomClient(BaseDicomClient):
    def __init__(self, dicom_dir: str, dicom_source_dir: str, *args, **kwargs) -> None:
        """
        :param dicom_dir: the DICOM output dir for image retrievals (same as other clients)
        :param dicom_source_dir: source directory for *.dcm files
        """
        self.dicom_dir = dicom_dir
        os.makedirs(self.dicom_dir, exist_ok=True)
        self.dicom_source_dir = dicom_source_dir

        self.dicom_datasets: Dict[str, Dataset] = {}

        for dicom_file in glob.glob(f'{dicom_source_dir}/**/*.dcm', recursive=True):
            self._read_and_add_data_set(dicom_file)

    def _read_and_add_data_set(self, filename: str) -> None:
        filepath = self._filepath(filename)
        self._add_dataset(dcmread(filepath, stop_before_pixels=True), filepath)

    def _add_dataset(self, dataset: Dataset, filepath: str = None) -> None:
        if filepath is None:
            filepath = self._filepath(dicom_filename(dataset))
        self.dicom_datasets[filepath] = dataset

    def _filepath(self, filename):
        return os.path.join(self.dicom_source_dir, filename)

    def verify(self) -> bool:
        return True

    def search_patients(self, search_query: Optional[str] = None,
                        search_query_type: Optional[str] = None,
                        additional_tags: Optional[List[str]] = None,
                        wildcard: bool = True) -> List[Dataset]:
        '''
        :param search_query: String containing query value to find.
            PatientID, PatientName, or, if wildcard is True, any partial there of.
        :param search_query_type: Optional string that restricts patient search to 'PatientID' or 'PatientName'.
            If None, perform a find for the query once on PatientID and again on PatientName.
        :param additional_tags: Additional dicom attributes that should be
            included in the returned patient dicom dataset values object.
        :param wildcard: Boolean stating whether to search based on
            any PatientName or PatientID partial string (i.e. Sam would find Samuel).
        :returns: List of DICOM query responses for each patient matching the query.
        '''
        patient_id_to_results = defaultdict(Dataset)

        # Build patient-level datasets from the instance-level test data
        for dataset in self.dicom_datasets.values():
            patient_id = getattr(dataset, 'PatientID', '').lower()
            patient_name = str(getattr(dataset, 'PatientName', '')).lower()
            search_query = search_query.lower()
            if wildcard:
                match = (search_query in patient_id) or (search_query in patient_name)
            else:
                if search_query_type == 'PatientID' or search_query_type is None:
                    match = (search_query == patient_id)
                if search_query_type == 'PatientName' or search_query_type is None:
                    match = (search_query == patient_name)
            if match:
                result = patient_id_to_results[patient_id]
                self.update_patient_result(result, dataset, additional_tags)
        return list(patient_id_to_results.values())

    def search_series(self, query_dataset, additional_tags=None) -> List[Dataset]:
        # Build series-level datasets from the instance-level test data
        additional_tags = additional_tags or []
        result_datasets = []
        for dataset in self.dicom_datasets.values():
            series_matches = dataset.SeriesInstanceUID == query_dataset.SeriesInstanceUID
            if series_matches:
                ds = Dataset()
                additional_tags += [
                    'PatientName',
                    'PatientBirthDate',
                    'BodyPartExamined',
                    'SeriesDescription',
                    'PatientPosition',
                ]
                ds.PatientStudyInstanceUIDs = MultiValue(UID, [dataset.StudyInstanceUID])
                ds.PacsmanPrivateIdentifier = PRIVATE_ID
                ds.PatientMostRecentStudyDate = getattr(dataset, 'StudyDate', '')
                copy_dicom_attributes(ds, dataset, additional_tags)
                result_datasets.append(ds)
        return result_datasets

    def studies_for_patient(self, patient_id, study_date_tag=None, additional_tags=None) -> List[Dataset]:
        # additional tags are ignored here; only tags available are already in the files
        study_id_to_dataset: Dict[str, Dataset] = {}

        date_format_str = '%Y%m%d'  # e.g. 20210101
        study_start_date = study_end_date = None
        if study_date_tag is not None:
            raw_study_start_str, raw_study_end_str = study_date_tag.split('-')
            # Accepted formats for tag are `START-END`, `START-` and `-END`
            if raw_study_start_str:
                study_start_date = datetime.strptime(raw_study_start_str, date_format_str).date()
            if raw_study_end_str:
                study_end_date = datetime.strptime(raw_study_end_str, date_format_str).date()

        def date_filter(study_ds):
            if hasattr(study_ds, 'StudyDate'):
                study_date_str = dataset.StudyDate
            elif hasattr(study_ds, 'SeriesDate'):
                study_date_str = dataset.SeriesDate
            else:
                study_date_str = None

            if (study_start_date is None and study_end_date is None) or study_date_str is None:
                return True

            study_date = datetime.strptime(study_date_str, date_format_str).date()
            if study_start_date and study_end_date:
                return study_date >= study_start_date and study_date <= study_end_date
            elif study_start_date:
                return study_date >= study_start_date
            else:
                # Should already be covered by edge-case early return
                assert study_end_date is not None
                return study_date <= study_end_date

        # Return one dataset per study
        for dataset in self.dicom_datasets.values():
            if patient_id == dataset.PatientID and dataset.StudyInstanceUID not in study_id_to_dataset:
                if date_filter(dataset):
                    study_id_to_dataset[dataset.StudyInstanceUID] = dataset
        return list(study_id_to_dataset.values())

    def series_for_study(self, study_id, modality_filter=None, additional_tags=None,
                         manual_count=True) -> List[Dataset]:
        # Build series-level datasets from the instance-level test data
        series_id_to_dataset: Dict[str, Dataset] = {}
        for dataset in self.dicom_datasets.values():
            study_matches = dataset.StudyInstanceUID == study_id
            modality_matches = modality_filter is None or getattr(dataset, 'Modality', '') in modality_filter
            if study_matches and modality_matches:
                dataset.PacsmanPrivateIdentifier = PRIVATE_ID
                dataset.BodyPartExamined = getattr(dataset, 'BodyPartExamined', '')
                dataset.SeriesDescription = getattr(dataset, 'SeriesDescription', '')
                dataset.PatientPosition = getattr(dataset, 'PatientPosition', '')
                series_id = dataset.SeriesInstanceUID
                if series_id in series_id_to_dataset:
                    series_id_to_dataset[series_id].NumberOfSeriesRelatedInstances += 1
                else:
                    dataset.NumberOfSeriesRelatedInstances = 1
                    series_id_to_dataset[series_id] = dataset

        return list(series_id_to_dataset.values())

    def images_for_series(self, study_id, series_id, additional_tags=None, max_count=None) -> List[Dataset]:
        image_datasets = []
        for dataset in self.dicom_datasets.values():
            series_matches = dataset.SeriesInstanceUID == series_id and dataset.StudyInstanceUID == study_id
            if series_matches:
                image_datasets.append(dataset)
            if max_count and len(image_datasets) >= max_count:
                break
        return image_datasets

    def fetch_images_as_dicom_files(self, study_id: str, series_id: str) -> Optional[str]:
        result_dir = os.path.join(self.dicom_dir, series_id)
        os.makedirs(result_dir, exist_ok=True)
        found = False
        for (path, ds) in self.dicom_datasets.items():
            if ds.SeriesInstanceUID == series_id:
                found = True
                dest_path = os.path.join(result_dir, f'{ds.SOPInstanceUID}.dcm')
                shutil.copy(path, dest_path)
        if found:
            return result_dir
        else:
            return None

    def fetch_image_as_dicom_file(self, study_id: str, series_id: str, sop_instance_id: str) -> Optional[str]:
        result_dir = os.path.join(self.dicom_dir, series_id)
        os.makedirs(result_dir, exist_ok=True)
        for (path, ds) in self.dicom_datasets.items():
            if ds.SOPInstanceUID == sop_instance_id:
                return shutil.copy(path, os.path.join(result_dir))
        return None

    def fetch_thumbnail(self, study_id: str, series_id: str) -> Optional[str]:
        series_items = []
        for path_to_ds in self.dicom_datasets.items():
            if path_to_ds[1].SeriesInstanceUID == series_id:
                series_items.append(path_to_ds)
        if not series_items:
            return None

        series_items = sorted(series_items, key=lambda t: t[1].SOPInstanceUID)

        thumbnail_index = len(series_items) // 2
        thumbnail_archive_path = series_items[thumbnail_index][0]
        thumbnail_instance_id = series_items[thumbnail_index][1].SOPInstanceUID

        # copying to instance ID ensures that the filename is unique
        dcm_path = os.path.join(self.dicom_dir, f'{thumbnail_instance_id}.dcm')
        shutil.copy(thumbnail_archive_path, dcm_path)

        png_path = process_and_write_png_from_file(dcm_path)
        return png_path

    def fetch_slice_thumbnail(self, study_id: str, series_id: str,
                              instance_id: str) -> Optional[str]:
        for path, ds in self.dicom_datasets.items():
            if ds.SeriesInstanceUID == series_id and ds.SOPInstanceUID == instance_id:
                thumbnail_series_path = path
                dcm_path = os.path.join(self.dicom_dir, f'{instance_id}.dcm')
                shutil.copy(thumbnail_series_path, dcm_path)
                png_path = process_and_write_png_from_file(dcm_path)
                return png_path
        logger.warning(f'Could not find instance {instance_id} for series {series_id}')
        return None

    def send_datasets(self, datasets: Iterable[Dataset], override_remote_ae: str = None,
                      override_pacs_url: str = None, override_pacs_port: int = None) -> None:
        new_dicom_datasets = {}
        for dataset in datasets:
            filepath = self._filepath(dicom_filename(dataset))
            new_dicom_datasets[filepath] = dataset
        self.dicom_datasets = {**self.dicom_datasets, **new_dicom_datasets}



'''
Remote tests depend on a running Orthanc instance (emulating a remote server).
No data is fetched from the remote server because C-GET is not yet supported in pacsman.

Local tests depend on data in the `test_dicom_dir` directory, which is imported into a local
Horos instance.

Steps to run integration tests:

 1) Configure Horos listener to run on port 11112.
 2*) In Horos: File -> Import -> Import Files and import all files in the 'test_dicom_data' dir
 3) Add localhost:11113 as a location in Horos with AETitle "TEST-SCP"
 4) `pytest integration_test.py` or `pytest -m local integration_test.py` for local-only

An alternative to step 2 above is to run upload_test.py

If horos is running on a different machine set the LOCAL_PACS_URL environment variable to
the ip of the machine running horos and similarly replace localhost in step
3 of the instructions above.

'''

import logging
import os
from unittest import mock

import pytest

from .base_client import BaseDicomClient
from .filesystem_dev_client import FilesystemDicomClient
from .pynetdicom_client import PynetDicomClient
from .dcmtk_client import DcmtkDicomClient


def initialize_pynetdicom_client(client_ae, pacs_url, pacs_port, dicom_dir):
    return PynetDicomClient(client_ae=client_ae, remote_ae='ORTHANC', pacs_url=pacs_url, pacs_port=pacs_port,
                            dicom_dir=dicom_dir)


def initialize_filesystem_client(dicom_dir, *args, **kwargs):
    file_dir = os.path.dirname(os.path.abspath(__file__))
    dicom_source_dir = os.path.join(file_dir, 'test_dicom_data')
    return FilesystemDicomClient(dicom_dir=dicom_dir, dicom_source_dir=dicom_source_dir,
                                 client_ae="asdf")


def initialize_dcmtk_client(client_ae, pacs_url, pacs_port, dicom_dir):
    return DcmtkDicomClient(client_ae=client_ae,
                            remote_ae='ORTHANC',
                            pacs_url=pacs_url,
                            pacs_port=pacs_port,
                            dicom_dir=dicom_dir)


dicom_client_initializers = [initialize_pynetdicom_client, initialize_filesystem_client]

LOCAL_PACS_URL = os.environ.get('LOCAL_PACS_URL', 'localhost')


@pytest.fixture(scope="module", params=dicom_client_initializers)
def local_client(request):
    logger = logging.getLogger(str(request.param))
    stream_logger = logging.StreamHandler()
    logger.addHandler(stream_logger)
    logger.setLevel(logging.DEBUG)
    pynetdicom_logger = logging.getLogger('pynetdicom3')
    pynetdicom_logger.setLevel(logging.DEBUG)
    # local (Horos, all PAT014 data pulled from dicomserver.co.uk)
    return request.param(client_ae='TEST', pacs_url=LOCAL_PACS_URL,
                         pacs_port=11112, dicom_dir='.')


@pytest.fixture(scope="module", params=dicom_client_initializers)
def remote_client(request):
    logger = logging.getLogger(str(request.param))
    stream_logger = logging.StreamHandler()
    logger.addHandler(stream_logger)
    logger.setLevel(logging.DEBUG)
    pynetdicom_logger = logging.getLogger('pynetdicom3')
    pynetdicom_logger.setLevel(logging.DEBUG)
    return request.param(client_ae='TEST', pacs_url='localhost',
                         pacs_port=4242, dicom_dir='.')


@pytest.fixture(scope="module")
def dcmtk_client():
    logger = logging.getLogger('dcmtk_logger')
    stream_logger = logging.StreamHandler()
    logger.addHandler(stream_logger)
    logger.setLevel(logging.DEBUG)
    return initialize_dcmtk_client(client_ae='TEST', pacs_url='localhost',
                                   pacs_port=4242, dicom_dir='.')


@pytest.fixture(scope="function")
def c_find_mock(dcmtk_client):
    with mock.patch.object(dcmtk_client, '_send_c_find', wraps=dcmtk_client._send_c_find) as patched_cfind:
        yield patched_cfind


@pytest.mark.integration
@pytest.mark.local
def test_verify_c_echo(local_client):
    assert local_client.verify()


@pytest.mark.integration
@pytest.mark.remote
def test_local_patient_search(dcmtk_client, c_find_mock):
    patient_datasets = dcmtk_client.search_patients(search_query='PAT014',
                                                    additional_tags=['PatientSex'])
    assert len(patient_datasets) == 1
    assert len(patient_datasets[0].PatientStudyInstanceUIDs) > 1
    assert patient_datasets[0].PatientMostRecentStudyDate
    assert patient_datasets[0].PatientSex == 'F'
    # assert c_find got called 2x, once for PatientID and once for PatientName
    assert c_find_mock.call_count == 2

    # Patient name search should also succeed by default
    patient_datasets = dcmtk_client.search_patients(
        search_query='Richardson',
        wildcard=True
    )
    assert len(patient_datasets) == 1

    c_find_mock.reset_mock()

    patient_datasets = dcmtk_client.search_patients(
        search_query='PAT014',
        search_query_type='PatientID',
        wildcard=False,
    )
    assert len(patient_datasets) == 1
    # assert c_find only got called 1x
    c_find_mock.assert_called_once()
    c_find_mock.reset_mock()

    # Patient name should not return results for PAT014
    patient_datasets = dcmtk_client.search_patients(
        search_query='PAT014',
        search_query_type='PatientName',
        wildcard=False,
    )
    # assert c_find only got called 1x
    c_find_mock.assert_called_once()
    assert len(patient_datasets) == 0

    # Patient name search should succeed
    patient_datasets = dcmtk_client.search_patients(
        search_query='Richardson',
        search_query_type='PatientName',
        wildcard=True,
    )
    assert len(patient_datasets) == 1


@pytest.mark.integration
@pytest.mark.local
def test_local_patient_search_wildcard(local_client):
    patient_datasets = local_client.search_patients('*',
                                                    additional_tags=['PatientSex'])
    assert len(patient_datasets) > 1


@pytest.mark.integration
@pytest.mark.local
def test_local_series_for_study(local_client):
    # this series is for patient PAT014
    series_datasets = local_client.series_for_study('1.2.826.0.1.3680043.11.118',
                                                    modality_filter=['CT'],
                                                    additional_tags=['InstitutionName'])
    assert len(series_datasets) > 1
    assert series_datasets[0]
    for ds in series_datasets:
        assert ds.NumberOfSeriesRelatedInstances >= 1
        assert ds.InstitutionName


@pytest.mark.integration
@pytest.mark.local
def test_local_studies_for_patient(local_client):
    studies_datasets = local_client.studies_for_patient('PAT014')

    assert len(studies_datasets) > 1
    assert studies_datasets[0]
    for ds in studies_datasets:
        assert ds.StudyInstanceUID


@pytest.mark.integration
@pytest.mark.remote
def test_local_studies_date_filtering(remote_client: BaseDicomClient):
    # Patient has two datasets, with dates `20180521` and `20180522`
    PATIENT = 'PAT014'

    # No date filter
    series_datasets = remote_client.studies_for_patient(PATIENT)
    assert len(series_datasets) == 2

    # START-END
    series_datasets = remote_client.studies_for_patient(PATIENT, study_date_tag='20180501-20180521')
    assert len(series_datasets) == 1

    # START-
    series_datasets = remote_client.studies_for_patient(PATIENT, study_date_tag='20180501-')
    assert len(series_datasets) == 2

    # -END
    series_datasets = remote_client.studies_for_patient(PATIENT, study_date_tag='-20180521')
    assert len(series_datasets) == 1


@pytest.mark.integration
@pytest.mark.local
def test_local_fetch(local_client, tmpdir):
    series_id = '1.2.826.0.1.3680043.6.51581.36765.20180518132103.25992.1.21'
    local_client.dicom_dir = str(tmpdir)
    local_client.fetch_images_as_dicom_files(series_id)

    series_dir = os.path.join(tmpdir, series_id)
    assert os.path.isdir(series_dir)
    assert len(os.listdir(series_dir)) > 1


@pytest.mark.integration
@pytest.mark.local
def test_local_fetch_thumbnail(local_client, tmpdir):
    # Patient ID E3148
    series_id = '1.2.392.200193.3.1626980217.161129.153348.41538611151089740341'
    local_client.dicom_dir = tmpdir
    thumbnail_path = local_client.fetch_thumbnail(series_id)
    assert thumbnail_path
    assert len(os.listdir(tmpdir)) == 1


@pytest.mark.integration
@pytest.mark.local
def test_local_fetch_fail(local_client, tmpdir):
    series_id = 'nonexistentseriesID'
    local_client.dicom_dir = tmpdir
    result_dir = local_client.fetch_images_as_dicom_files(series_id)
    thumbnail_file = local_client.fetch_thumbnail(series_id)
    assert result_dir is None
    assert thumbnail_file is None


@pytest.mark.integration
@pytest.mark.remote
def test_verify_c_echo_remote(remote_client):
    assert remote_client.verify()


@pytest.mark.integration
@pytest.mark.remote
def test_remote_patient_search(remote_client):
    patient_datasets = remote_client.search_patients('PAT014')
    assert len(patient_datasets) >= 1
    for ds in patient_datasets:
        assert ds.PatientID == 'PAT014'
        assert ds.PatientMostRecentStudyDate
        assert ds.PatientStudyInstanceUIDs


@pytest.mark.integration
@pytest.mark.remote
@pytest.mark.skip(reason="data is missing on remote server")
def test_remote_series_for_study(remote_client):
    # this series is for patient PAT014
    series_datasets = remote_client.series_for_study('1.2.826.0.1.3680043.11.118')
    assert len(series_datasets) > 1


@pytest.mark.integration
@pytest.mark.remote
def test_remote_fetch_fail(remote_client):
    # Skip failure check for dummy client (which never fails)
    if isinstance(remote_client, FilesystemDicomClient):
        return

    # on dicomserver.co.uk, fails with 'Unknown Move Destination: TEST-SCP'
    with pytest.raises(Exception):
        remote_client.fetch_images_as_dicom_files('1.2.826.0.1.3680043.6.79369.13951.20180518132058.25992.1.15')
